# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import shutil
import tarfile
import subprocess

import six
if six.PY3:
    from urllib.request import (
        urlretrieve
    )
else:
    from urllib import urlretrieve

from base64 import b64decode

from charmhelpers.core.host import (
    mkdir,
    write_file,
    service_restart,
)

from charmhelpers.contrib.openstack import context

from charmhelpers.contrib.hahelpers.cluster import (
    DC_RESOURCE_NAME,
    determine_apache_port,
    determine_api_port,
    is_elected_leader,
)

from charmhelpers.core.hookenv import (
    config,
    log,
    leader_get,
    DEBUG,
    INFO,
)

from charmhelpers.core.strutils import (
    bool_from_string,
)

from charmhelpers.contrib.hahelpers.apache import install_ca_cert

CA_CERT_PATH = '/usr/local/share/ca-certificates/keystone_juju_ca_cert.crt'


def is_cert_provided_in_config():
    ca = config('ssl_ca')
    cert = config('ssl_cert')
    key = config('ssl_key')
    return bool(ca and cert and key)


class ApacheSSLContext(context.ApacheSSLContext):

    interfaces = ['https']
    external_ports = []
    service_namespace = 'keystone'

    def __call__(self):
        # late import to work around circular dependency
        from keystone_utils import (
            determine_ports,
            update_hash_from_path,
        )

        ssl_paths = [CA_CERT_PATH,
                     os.path.join('/etc/apache2/ssl/',
                                  self.service_namespace)]

        self.external_ports = determine_ports()
        before = hashlib.sha256()
        for path in ssl_paths:
            update_hash_from_path(before, path)

        ret = super(ApacheSSLContext, self).__call__()

        after = hashlib.sha256()
        for path in ssl_paths:
            update_hash_from_path(after, path)

        # Ensure that apache2 is restarted if these change
        if before.hexdigest() != after.hexdigest():
            service_restart('apache2')

        return ret

    def configure_cert(self, cn):
        from keystone_utils import (
            SSH_USER,
            get_ca,
            ensure_permissions,
            is_ssl_cert_master,
        )

        # Ensure ssl dir exists whether master or not
        ssl_dir = os.path.join('/etc/apache2/ssl/', self.service_namespace)
        perms = 0o755
        mkdir(path=ssl_dir, owner=SSH_USER, group='keystone', perms=perms)
        # Ensure accessible by keystone ssh user and group (for sync)
        ensure_permissions(ssl_dir, user=SSH_USER, group='keystone',
                           perms=perms)

        if not is_cert_provided_in_config() and not is_ssl_cert_master():
            log("Not ssl-cert-master - skipping apache cert config until "
                "master is elected", level=INFO)
            return

        log("Creating apache ssl certs in %s" % (ssl_dir), level=INFO)

        cert = config('ssl_cert')
        key = config('ssl_key')

        if not (cert and key):
            ca = get_ca(user=SSH_USER)
            cert, key = ca.get_cert_and_key(common_name=cn)
        else:
            cert = b64decode(cert)
            key = b64decode(key)

        write_file(path=os.path.join(ssl_dir, 'cert_{}'.format(cn)),
                   content=cert, owner=SSH_USER, group='keystone', perms=0o644)
        write_file(path=os.path.join(ssl_dir, 'key_{}'.format(cn)),
                   content=key, owner=SSH_USER, group='keystone', perms=0o644)

    def configure_ca(self):
        from keystone_utils import (
            SSH_USER,
            get_ca,
            ensure_permissions,
            is_ssl_cert_master,
        )

        if not is_cert_provided_in_config() and not is_ssl_cert_master():
            log("Not ssl-cert-master - skipping apache ca config until "
                "master is elected", level=INFO)
            return

        ca_cert = config('ssl_ca')
        if ca_cert is None:
            ca = get_ca(user=SSH_USER)
            ca_cert = ca.get_ca_bundle()
        else:
            ca_cert = b64decode(ca_cert)

        # Ensure accessible by keystone ssh user and group (unison)
        install_ca_cert(ca_cert)
        ensure_permissions(CA_CERT_PATH, user=SSH_USER, group='keystone',
                           perms=0o0644)

    def canonical_names(self):
        addresses = self.get_network_addresses()
        addrs = []
        for address, endpoint in addresses:
            addrs.append(endpoint)

        return list(set(addrs))


class HAProxyContext(context.HAProxyContext):
    interfaces = []

    def __call__(self):
        '''
        Extends the main charmhelpers HAProxyContext with a port mapping
        specific to this charm.
        Also used to extend nova.conf context with correct api_listening_ports
        '''
        from keystone_utils import api_port
        ctxt = super(HAProxyContext, self).__call__()

        # determine which port api processes should bind to, depending
        # on existence of haproxy + apache frontends
        listen_ports = {}
        listen_ports['admin_port'] = api_port('keystone-admin')
        listen_ports['public_port'] = api_port('keystone-public')

        # Apache ports
        a_admin_port = determine_apache_port(api_port('keystone-admin'),
                                             singlenode_mode=True)
        a_public_port = determine_apache_port(api_port('keystone-public'),
                                              singlenode_mode=True)

        port_mapping = {
            'admin-port': [
                api_port('keystone-admin'), a_admin_port],
            'public-port': [
                api_port('keystone-public'), a_public_port],
        }

        # for haproxy.conf
        ctxt['service_ports'] = port_mapping
        # for keystone.conf
        ctxt['listen_ports'] = listen_ports
        return ctxt


class KeystoneContext(context.OSContextGenerator):
    interfaces = []

    def __call__(self):
        from keystone_utils import (
            api_port, set_admin_token, endpoint_url, resolve_address,
            PUBLIC, ADMIN, PKI_CERTS_DIR, ensure_pki_cert_paths, ADMIN_DOMAIN,
        )
        ctxt = {}
        ctxt['token'] = set_admin_token(config('admin-token'))
        ctxt['api_version'] = int(config('preferred-api-version'))
        ctxt['admin_role'] = config('admin-role')
        if ctxt['api_version'] > 2:
            ctxt['service_tenant_id'] = \
                leader_get(attribute='service_tenant_id')
            ctxt['admin_domain_name'] = ADMIN_DOMAIN
            ctxt['admin_domain_id'] = \
                leader_get(attribute='admin_domain_id')
            ctxt['default_domain_id'] = \
                leader_get(attribute='default_domain_id')
        ctxt['admin_port'] = determine_api_port(api_port('keystone-admin'),
                                                singlenode_mode=True)
        ctxt['public_port'] = determine_api_port(api_port('keystone-public'),
                                                 singlenode_mode=True)

        ctxt['debug'] = config('debug')
        ctxt['verbose'] = config('verbose')
        ctxt['token_expiration'] = config('token-expiration')

        ctxt['identity_backend'] = config('identity-backend')
        ctxt['assignment_backend'] = config('assignment-backend')
        if config('identity-backend') == 'ldap':
            ctxt['ldap_server'] = config('ldap-server')
            ctxt['ldap_user'] = config('ldap-user')
            ctxt['ldap_password'] = config('ldap-password')
            ctxt['ldap_suffix'] = config('ldap-suffix')
            ctxt['ldap_readonly'] = config('ldap-readonly')
            ldap_flags = config('ldap-config-flags')
            if ldap_flags:
                flags = context.config_flags_parser(ldap_flags)
                ctxt['ldap_config_flags'] = flags

        ctxt['enable_saml2'] = config('enable-saml2')
        ctxt['enable_oidc'] = config('enable-oidc')
        if (ctxt['enable_saml2'] or ctxt['enable_oidc']):
            ctxt['trusted_dashboard'] = config('trusted-dashboard')
            ctxt['trusted_dashboard_auth_root'] = '/auth/websso' \
                if not config('trusted-dashboard-auth-root') \
                else config('trusted-dashboard-auth-root')
            ctxt['trusted_dashboard_auth_full'] = \
                os.path.join(ctxt['trusted_dashboard'],
                             ctxt['trusted_dashboard_auth_root'], '')
        if ctxt['enable_saml2']:
            ctxt['saml2_metadata_url'] = config('saml2-metadata-url')
            ctxt['saml2_metadata_bck_file'] = \
                'federation_metadata.xml'
            ctxt['saml2_metadata_signer_url'] = \
                config('saml2-metadata-signer-url')
            ctxt['saml2_metadata_signer_bck_file'] = \
                'federation_metadata_signer.pem'
            ctxt['saml2_fed_users_domain'] = 'cloudusers' \
                if not config('saml2-fed-users-domain') \
                else config('saml2-fed-users-domain')
            ctxt['saml2_mapping'] = config('saml2-mapping')
            ctxt['shibsp_discovery_enabled'] = False
            ctxt['shibsp_discovery_edsconfig_url'] = \
                config('shibsp-discovery-edsconfig-url')
            if config('shibsp-discovery-edsconfig-url'):
                ctxt['shibsp_discovery_enabled'] = True
                ctxt['shibsp_eds_url'] = endpoint_url(
                    resolve_address(PUBLIC),
                    api_port('keystone-public')) + '/eds'
            ctxt['shibsp_application_id'] = config('shibsp-application-id')
            ctxt['shibsp_idp_enabled'] = False
            ctxt['shibsp_idp_remoteids'] = config('shibsp-idp-remoteids')
            if config('shibsp-idp-hostname'):
                ctxt['shibsp_idp_enabled'] = True
                ctxt['shibsp_idp_hostname'] = config('shibsp-idp-hostname')
                ctxt['shibsp_idp_entity_id'] = \
                    'https://' + config('shibsp-idp-hostname') + \
                    '/idp/shibboleth' \
                    if not config('shibsp-idp-entity-id') \
                    else config('shibsp-idp-entity-id')
                ctxt['shibsp_idp_metadata_url'] = \
                    config('shibsp-idp-metadata-url')
                ctxt['shibsp_idp_metadata_file'] = \
                    config('shibsp-idp-metadata-file')
                ctxt['shibsp_idp_metadata_bck_file'] = \
                    config('shibsp-idp-hostname') + '-metadata.xml'

            ctxt['shibsp_support_contact'] = config('shibsp-support-contact')
            ctxt['shibsp_check_eppn_scope'] = 0
        if ctxt['enable_oidc']:
            ctxt['oidc_claimprefix'] = '"OIDC-"'
            ctxt['oidc_responsetype'] = '"id_token"'
            ctxt['oidc_scope'] = '"openid email profile"'
            ctxt['oidc_provider_metadata_url'] = \
                config('oidc-provider-metadata-url')
            ctxt['oidc_client_id'] = config('oidc-client-id')
            ctxt['oidc_client_secret'] = config('oidc-client-secret')
            ctxt['oidc_crypto_passphrase'] = \
                config('oidc-crypto-passphrase')
            ctxt['oidc_op_id'] = config('oidc-op-id')
            ctxt['oidc_op_remoteids'] = config('oidc-op-remoteids')
            ctxt['oidc_fed_users_domain'] = 'cloudusers' \
                if not config('oidc-fed-users-domain') \
                else config('oidc-fed-users-domain')
            ctxt['oidc_mapping'] = config('oidc-mapping')
            ctxt['oidc_redirect_uri'] = endpoint_url(
                resolve_address(PUBLIC), api_port('keystone-public'),
                '/v3/auth/OS-FEDERATION/websso/oidc/redirect') \
                if not config('oidc-redirect-uri') \
                else config('oidc-redirect-uri')

        enable_pki = config('enable-pki')
        if enable_pki and bool_from_string(enable_pki):
            log("Enabling PKI", level=DEBUG)
            ctxt['token_provider'] = 'pki'

        ensure_pki_cert_paths()
        certs = os.path.join(PKI_CERTS_DIR, 'certs')
        privates = os.path.join(PKI_CERTS_DIR, 'privates')
        ctxt.update({'certfile': os.path.join(certs, 'signing_cert.pem'),
                     'keyfile': os.path.join(privates, 'signing_key.pem'),
                     'ca_certs': os.path.join(certs, 'ca.pem'),
                     'ca_key': os.path.join(certs, 'ca_key.pem')})

        # Base endpoint URL's which are used in keystone responses
        # to unauthenticated requests to redirect clients to the
        # correct auth URL.
        ctxt['public_endpoint'] = endpoint_url(
            resolve_address(PUBLIC),
            api_port('keystone-public')).replace('v2.0', '')
        ctxt['admin_endpoint'] = endpoint_url(
            resolve_address(ADMIN),
            api_port('keystone-admin')).replace('v2.0', '')

        if ctxt['enable_saml2']:
            ctxt['shibsp_entity_id'] = endpoint_url(
                resolve_address(PUBLIC),
                api_port('keystone-public')) + '/shibboleth'
            self.configure_saml2(ctxt)
        return ctxt

    def configure_saml2(self, ctxt):
        from keystone_utils import (
            ensure_permissions,
        )

        log("SAML2 configuration enabled, create metadata dir", level=INFO)
        shib_dir = '/etc/shibboleth/'
        shib_meta_dir = os.path.join(shib_dir, 'metadata')
        perms = 0o755
        mkdir(path=shib_dir, owner='_shibd', group='_shibd', perms=perms)
        mkdir(path=shib_meta_dir, owner='_shibd', group='_shibd',
              perms=perms)
        # Ensure accessible by '_shibd' user and group
        ensure_permissions(shib_dir, user='_shibd', group='_shibd',
                           perms=perms)
        ensure_permissions(shib_meta_dir, user='_shibd', group='_shibd',
                           perms=perms)

        if ctxt['shibsp_discovery_enabled']:
            log("SAML2 discovery service configuration download ", level=INFO)
            eds_config_url = ctxt['shibsp_discovery_edsconfig_url']
            eds_config_destdir = '/var/www/html/eds'
            if not os.path.exists(eds_config_destdir):
                mkdir(eds_config_destdir, perms=0o755)
            eds_config_tarfile = \
                os.path.join('/tmp', 'eds_config.tar.gz')
            try:
                urlretrieve(eds_config_url, eds_config_tarfile)
            except:
                log("SAML2 discovery service configuration URL -->" +
                    eds_config_url +
                    "<-- undefined or unreachable", level=INFO)

            edsConfigH = tarfile.open(eds_config_tarfile)
            edsConfigFiles = edsConfigH.getmembers()
            edsConfigH.extractall(eds_config_destdir)
            if edsConfigFiles[0].isdir():
                edsConfigSubdir = \
                    os.path.join(eds_config_destdir, edsConfigFiles[0].name)
                # Archive starts with subdir, move everything up to destdir
                for aFile in edsConfigFiles:
                    if aFile == edsConfigFiles[0]:
                        continue
                    aFileFullname = \
                        os.path.join(eds_config_destdir, aFile.name)
                    shutil.copy2(aFileFullname, eds_config_destdir)
                shutil.rmtree(edsConfigSubdir)
            edsConfigH.close()
            if config('shibsp-discovery-edsconfig-script'):
                eds_config_scriptname = \
                    config('shibsp-discovery-edsconfig-script')
                eds_config_script = \
                    os.path.join(eds_config_destdir, eds_config_scriptname)
                if os.path.isfile(eds_config_script) and \
                   os.access(eds_config_script, os.X_OK):
                    # File exists and is executable
                    subprocess.call([eds_config_script,
                                     ctxt['trusted_dashboard']])

        log("SAML2 federation metadata download ", level=INFO)
        fed_metadata_signer_url = ctxt['saml2_metadata_signer_url']
        fed_metadata_signer_pem = \
            os.path.join(shib_dir, 'metadata',
                         ctxt['saml2_metadata_signer_bck_file'])
        try:
            urlretrieve(fed_metadata_signer_url, fed_metadata_signer_pem)
            ensure_permissions(fed_metadata_signer_pem,
                               user='_shibd', group='_shibd', perms=0o644)
        except:
            log("SAML2 federation metadata URL -->" + fed_metadata_signer_url +
                "<-- undefined or unreachable", level=INFO)

        if ctxt['shibsp_idp_enabled']:
            log("Shibsp Idp metadata configuration", level=INFO)
            idp_metadata_bck_file = \
                os.path.join(shib_dir, 'metadata',
                             ctxt['shibsp_idp_metadata_bck_file'])
            if ctxt['shibsp_idp_metadata_file']:
                idp_metadata_file = ctxt['shibsp_idp_metadata_file']
                idp_metadata = b64decode(idp_metadata_file)
                write_file(path=idp_metadata_bck_file, content=idp_metadata,
                           owner='_shibd', group='_shibd', perms=0o644)
            elif ctxt['shibsp_idp_metadata_url']:
                idp_metadata_url = ctxt['shibsp_idp_metadata_url']
                try:
                    urlretrieve(idp_metadata_url, idp_metadata_bck_file)
                except:
                    log("Shibsp Idp metadata URL -->" + idp_metadata_url +
                        "<-- undefined or unreachable", level=INFO)

        # Shibboleth certificate
        log("Installing Shibboleth keys in %s" % (shib_dir), level=INFO)
        if not (config('shibsp-key') and config('shibsp-cert')):
            log("Undefined shibsp-key and/or shibsp-cert - "
                "skipping Shibboleth certificate config", level=INFO)
            return

        cert = config('shibsp-cert')
        key = config('shibsp-key')
        cert = b64decode(cert)
        key = b64decode(key)
        write_file(path=os.path.join(shib_dir, 'shibsp_cert.pem'),
                   content=cert, owner='_shibd', group='_shibd', perms=0o644)
        write_file(path=os.path.join(shib_dir, 'shibsp_key.pem'),
                   content=key, owner='_shibd', group='_shibd', perms=0o600)

        return


class KeystoneLoggingContext(context.OSContextGenerator):

    def __call__(self):
        ctxt = {}
        debug = config('debug')
        if debug:
            ctxt['root_level'] = 'DEBUG'
        log_level = config('log-level')
        log_level_accepted_params = ['WARNING', 'INFO', 'DEBUG', 'ERROR']
        if log_level in log_level_accepted_params:
            ctxt['log_level'] = config('log-level')
        else:
            log("log-level must be one of the following states "
                "(WARNING, INFO, DEBUG, ERROR) keeping the current state.")
            ctxt['log_level'] = None

        return ctxt


class TokenFlushContext(context.OSContextGenerator):

    def __call__(self):
        ctxt = {
            'token_flush': is_elected_leader(DC_RESOURCE_NAME)
        }
        return ctxt
